"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_1 = require("child_process");
const promises_1 = require("fs/promises");
const stream_1 = require("stream");
const ExecError_1 = __importDefault(require("./ExecError"));
const isReadable_1 = __importDefault(require("./isReadable"));
exports.default = (cmd, ...args) => {
    const env = Object.create(null);
    Object.assign(env, process.env);
    let cwd;
    let encoding = "utf8";
    let killOnStderr;
    let onStderr;
    let onStdout;
    let pidFile;
    let printCmdline = false;
    // printStd* are undefined by default to allow adaptive implicit config
    // depending on run method and onStd*.
    let printStderr;
    let printStdout;
    let stdin;
    // This is ignored if result type is status.
    let throwOnBadStatus = true;
    let throwOnSignal = true;
    let timeout;
    const run = (resultType, collectStdout, collectStderr) => {
        if (printCmdline) {
            console.debug("+", ...args);
        }
        const proc = (0, child_process_1.spawn)(cmd, args.map(String), {
            cwd,
            env,
            stdio: "pipe",
            timeout,
        });
        proc.once("spawn", () => {
            if (pidFile != undefined && proc.pid !== undefined) {
                // Attempt to write to file, but don't care if it fails.
                (0, promises_1.writeFile)(pidFile, proc.pid.toString()).catch(() => void 0);
            }
        });
        if (encoding != undefined) {
            proc.stdout.setEncoding(encoding);
            proc.stderr.setEncoding(encoding);
        }
        const resultStream = resultType == "stream" ? new stream_1.PassThrough() : undefined;
        const promise = new Promise((resolve, reject) => {
            if ((0, isReadable_1.default)(stdin)) {
                (0, stream_1.pipeline)(stdin, proc.stdin, (err) => {
                    if (err) {
                        // TODO Is any additional cleanup needed?
                        reject(err);
                    }
                });
            }
            else {
                if (stdin != undefined) {
                    proc.stdin.write(stdin);
                }
                proc.stdin.end();
            }
            const resultData = resultType == "data" ? Array() : undefined;
            proc.stderr.on("data", (chunk) => {
                if (printStderr) {
                    process.stderr.write(chunk);
                }
                onStderr?.(chunk);
                if (killOnStderr != undefined) {
                    proc.kill(killOnStderr);
                }
                if (collectStderr) {
                    resultStream?.write(chunk);
                    resultData?.push(chunk);
                }
            });
            proc.stdout.on("data", (chunk) => {
                if (printStdout) {
                    process.stdout.write(chunk);
                }
                onStdout?.(chunk);
                if (collectStdout) {
                    resultStream?.write(chunk);
                    resultData?.push(chunk);
                }
            });
            proc.on("error", reject);
            // WARNING: Listen to "close", not "exit". "exit" can be emitted before
            // stdio streams are closed, causing hard-to-debug race conditions where
            // stdout/stderr data is missing.
            proc.on("close", (code, signal) => {
                resultStream?.end();
                if (throwOnSignal && signal) {
                    reject(new ExecError_1.default(cmd, code ?? undefined, signal));
                }
                else if (resultType != "status" && throwOnBadStatus && code) {
                    reject(new ExecError_1.default(cmd, code, signal ?? undefined));
                }
                else if (resultData) {
                    resolve(encoding ? resultData.join("") : Buffer.concat(resultData));
                }
                else {
                    resolve(resultType != "none" ? code ?? -1 : undefined);
                }
            });
        });
        if (resultStream) {
            return {
                stream: resultStream,
                promise,
            };
        }
        return promise;
    };
    return {
        additionalEnv(entries) {
            for (const [k, v] of entries instanceof Map
                ? entries
                : Object.entries(entries)) {
                env[k] = v.toString();
            }
            return this;
        },
        binary() {
            encoding = undefined;
            return this;
        },
        env(name, value) {
            env[name] = value.toString();
            return this;
        },
        killOnStderr(signal = "SIGTERM") {
            killOnStderr =
                signal === false ? undefined : signal === true ? "SIGTERM" : signal;
            return this;
        },
        onStderr(onData) {
            onStderr = onData;
            printStderr ??= false;
            return this;
        },
        onStdout(onData) {
            onStdout = onData;
            printStdout ??= false;
            return this;
        },
        pidFile(path) {
            pidFile = path;
            return this;
        },
        print() {
            if (arguments.length == 1) {
                printStdout = printStderr = false;
            }
            else {
                [printStdout, printStderr] = arguments;
            }
            return this;
        },
        printCmdline(shouldPrintCmdline = true) {
            printCmdline = shouldPrintCmdline;
            return this;
        },
        printStderr(shouldPrint = true) {
            printStderr = shouldPrint;
            return this;
        },
        printStdout(shouldPrint = true) {
            printStdout = shouldPrint;
            return this;
        },
        stdin(data) {
            stdin = data;
            return this;
        },
        text(e = "utf-8") {
            encoding = e;
            return this;
        },
        throwOnBadStatus(shouldThrow = true) {
            throwOnBadStatus = shouldThrow;
            return this;
        },
        throwOnSignal(shouldThrow = true) {
            throwOnSignal = shouldThrow;
            return this;
        },
        timeout(ms) {
            timeout = ms;
            return this;
        },
        workingDir(dir) {
            cwd = dir;
            return this;
        },
        output(withStderr = false, withStdout = true) {
            printStdout ??= !withStdout;
            printStderr ??= !withStderr;
            return run("data", withStdout, withStderr);
        },
        run() {
            printStdout ??= true;
            printStderr ??= true;
            return run("none", false, false);
        },
        status() {
            printStdout ??= true;
            printStderr ??= true;
            return run("status", false, false);
        },
        stream(withStderr = false, withStdout = true) {
            printStdout ??= !withStdout;
            printStderr ??= !withStderr;
            return run("stream", withStdout, withStderr);
        },
    };
};
